#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022 "Under Mushroom Forest" July 1, 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};


//config
#define ITERATIONS 100
#define EPSILON 1e-2
#define EPSILON_SCALE 1.07
//#define DISTANCE_STEP 1.
#define STEP_DITHER .2732
#define VOLUME_EPSILON 1e-1
#define LIGHTING_EPSILON 1e-5

#define LIGHT_IMPORTANCE .5
#define LIGHT_DIRECTION normalize(vec3(.02973,1,-.2712))
#define LIGHT_SIZE .02
#define LIGHT_COLOR (vec3(.95,.85,.74))
#define LIGHT_RESCALE_MAX 1e5
#define LIGHT_CAUSTICS 100.
#define LIGHT_MASK(p) 1.

#define ANIMATION_TIME 60.
#define ANIMATION_SAMPLES 100
out float time;


//globals
#define PI 3.141592653589793

#ifdef ANIMATION_TIME
#define SFRAME (frame%ANIMATION_SAMPLES)
#else
#define SFRAME frame
#endif


//utility functions
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))

float twave(float x) {
	return abs(fract(x)*2.-1.);
}

float smoothMin(float a, float b, float r) {
    float l = clamp((a-b)*.5/r+.5, 0.,1.);
    return mix(a,b, l)-r*l*(1.-l);
}


vec2 mushrooms(vec3 p, vec2 seed) {
	vec3 rp = vec3(mod(abs(p.xz),40.)-20.,p.y).xzy;
	vec2 fp = floor(abs(p.xz)/40.);
	float lxz = length(rp.xz), height = 8.+twave(dot(fp,seed))*25.,
		gd = cos(atan(rp.x,rp.z)*30.)*.4*clamp(1.-abs(rp.y-(height-6.))*.2,0.,1.),//gills
		md = min(length(rp-vec3(0,2,0))-4.,max(rp.y-height,lxz-3.))-gd;//stems
	vec2 geo = vec2(md,1);
	
	float cd = max(abs(rp.y-height+lxz*.1)-max(-1.,4.-lxz*4./20.),lxz-20.)-gd;//caps
	if (cd < geo.x) geo = vec2(cd,p.y<(height-2.3)?1.:2.);
	
	return geo;
}


//scene definition

/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	/*animation*/
	float et = time;
	return mat3(vec3(0,1.6+cos(et*16.763)*.1,1)+normalize(vec3(.5,0,1))*et*10.,
			0,cos(et*.19734)*.3,.4+cos(et*.687624)*cos(et*.127683)*.6,
			1.,.5,0);
	/*screenshot 0
	return mat3(vec3(260,6,1080),
			0,.2,-4.1,
			1.,10,0);*/
	/*screenshot 1
	return mat3(vec3(2255,2,160),
			0,.8,3.5,
			.5,3,0);*/
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//plant/dirt ground
	vec2 geo;
	float gh = p.y;
	if (gh < .1) {
		vec3 mp = mod(abs(p-vec3(0,0.02,0)),.2)-.1;
		mp.xz *= r2d(p.x*2.973+p.y*3.8734+p.z*1.4763);
		mp.yz *= r2d(p.x*-1.9723+p.y*2.937+p.z*2.3873);
		mp = abs(mp)-.05;
		geo = vec2(max(gh,length(mp)-.04),0.);
	} else {
		geo = vec2(gh,0.);
	}
	
	//mushrooms
	vec2 md = mushrooms(p,vec2(.28763111,.7786211)),
		md2 = mushrooms(p+vec3(10,0,10),vec2(.457623,.5263));
	if (md.x < geo.x) geo = md;
	if (md2.x < geo.x) geo = md2;

	return geo;
}
vec2 geometryDetail(vec3 p) {
	return geometry(p);
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;*/
mat2x3 material(vec3 p, float mat) {
	if (mat == 2.) {
		//mushroom cap
		return mat2x3(vec3(.5,.02,.02), .8,0,1);
	}
	
	if (mat == 1.) {//mushroom stalk
		return mat2x3(vec3(.9,.82,.76), .14,0,1);
	}
	
	if (p.y < -.2) {//dirt
		return mat2x3(vec3(.4,.32,.26)*.4, .7,0,1);
	}
	//plant
	return mat2x3(vec3(.04,.45,.14), .24,0,1);
}


//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return mix(vec3(0),vec3(1),clamp((p.y-24.)/10.,0.,1.));
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	if (p.y < 34.) return vec3(0);
	return vec3(.45,.65,.9)*max(0.,d.y)*2.;
}


//sample detailed normal using derivatives from geometryDetail
vec3 normal(vec3 p) {
	vec3 n = normalize(vec3(geometryDetail(p+vec3(LIGHTING_EPSILON,0,0)).x,geometryDetail(p+vec3(0,LIGHTING_EPSILON,0)).x,geometryDetail(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometryDetail(p).x);
	if (n != n) return vec3(0,1,0);
	return n;
}


//bidirectional path tracer
uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION_TIME
	time = mod(float(frame)/float(framerate*ANIMATION_SAMPLES),ANIMATION_TIME);
	#else
	time = 0.;
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	vec4 random = hash(uvec4(SFRAME)*uvec4(1,10,100,1000))*I2F;
	vec3 rp = vec3(0,0,cam[2].z),
		rd = normalize(vec3(uv.xy+(random.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz+rd*cam[2].y;

	//path trace
	vec3 c = vec3(1);//ray color tint
	vec4 l = vec4(0);//l.xyz = light color accumulation, l.w = number of bounces or -1 if important ray
	float epsilon = EPSILON;
	for (int i = 0; i < ITERATIONS && rp.y < 34.; i++) {
		vec2 geo = geometry(rp);
		if (geo.x <= 0.) {//hit/inside geometry
			mat2x3 mat = material(rp,geo.y);//sample material
			if (random.w < mat[1].z) {//stochastic opacity
				
				#ifdef LIGHT_IMPORTANCE
				if (l.w < 0.) {	//failed importance cast
					c = vec3(0);
					break;
				}
				#endif
			
				rp -= rd*epsilon*2.;//back up ray
				c *= mat[0];//apply material color
				l += vec4(c.xyz*mat[1].y,1);//emissive lighting + bounce count
				random = hash(uvec4(abs(rp.xyz)*10000.,SFRAME))*I2F;
				vec3 nrm = normal(rp);

				#ifdef LIGHT_IMPORTANCE
				if (random.w < LIGHT_IMPORTANCE) {//directional light importance sampling
					//brdf
					float edge = (1.-LIGHT_SIZE*.1)-mat[1].x*4.,
						brdf = clamp((dot(LIGHT_DIRECTION,reflect(rd,nrm))-edge)/(1.-min(1.,edge)),0.,1.);
					rd = normalize(LIGHT_DIRECTION+(random.xyz-.5)*LIGHT_SIZE);
					
					if (dot(rd,nrm) < -.2 || brdf < 0.) {
						c = vec3(0);
						break;
					}
					c *= brdf/(1.+mat[1].x*LIGHT_CAUSTICS);
					if (l.w > 1.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w-1.));//rescale brightness of bounces
					l.w = -1.;//mark ray important
					
				} else {
				#endif
				
				//randomly bounce ray off microsurface bump
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
				float cs = pow(random.x,mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
				rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));
				#ifdef LIGHT_IMPORTANCE
				}
				#endif
				
				geo.x = 0.;
			} else {
				geo.x = VOLUME_EPSILON;
			}
		} else {
			#ifdef DISTANCE_STEP
			geo.x *= DISTANCE_STEP;
			#endif
		}
		
		//ray march
#ifdef STEP_DITHER
		float stepSz = (epsilon+geo.x)*((1.-STEP_DITHER)+random.w*STEP_DITHER);
#else
		float stepSz = epsilon+geo.x;
#endif
		rp += rd*stepSz;
		random.w = fract(random.w+stepSz*2763.861234);
#ifdef EPSILON_SCALE
		epsilon *= EPSILON_SCALE;
#endif
	}
	
	#ifdef LIGHT_IMPORTANCE
	if (l.w < 0.) {
		l.xyz += c*LIGHT_COLOR*(1.+LIGHT_CAUSTICS)*LIGHT_MASK(p)/LIGHT_IMPORTANCE;
	} else {
		if (l.w > 0.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w));//rescale brightness of bounces
	#endif
	//background
	l.xyz += c*(l.w>0.?backgroundLighting(rp,rd):background(rp,rd));
	#ifdef LIGHT_IMPORTANCE
	}
	#endif
	
	//sum light in buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (SFRAME == 0) sum = vec4(0);
	gl_FragColor = sum+vec4(l.xyz,1);
}
